(***********************************************************************)
(*                                                                     *)
(*                           Htmlc                                     *)
(*                                                                     *)
(*                    Pierre Weis, INRIA Rocquencourt                  *)
(*                                                                     *)
(*  Copyright 1997-2008,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: file.mli,v 1.2 2009/01/13 17:45:18 fclement Exp $ *)

(** {3 Operations on files.} *)

type file_extension = string
(* An alias to distinguish file name extensions from strings. *)
;;

type output_file_name = Path.file_name
(* An alias to distinguish output file names. *)
;;

type input_file_name = Path.file_name
(* An alias to distinguish input file names. *)
;;

(** {6 Naming.} *)

val name_out : output_file_name -> string
(** Returns the name of output file name. *)
;;

val name_in : input_file_name -> string
(** Returns the name of an input file name. *)
;;

(** {6 Getting and setting file permissions.} *)

val permissions : Path.file_name -> Unix.file_perm
(** Returns the file permission of a file name. *)
;;

val permissions_out : output_file_name -> Unix.file_perm
(** Returns the file permission of an output file name. *)
;;

val permissions_in : input_file_name -> Unix.file_perm
(** Returns the file permission of an input file name. *)
;;

val make_writable_file : Path.file_name -> unit
(** Turn a file writable at least for the owner.

   Land the current permission with 0o7577 to remove the writable flag to
   the owner, then add 0o200 to get the file writable by the owner.
   Note: if the file was already writable for others or the group, we keep
   this permission. *)
;;

val make_read_only_file : Path.file_name -> unit
(** Turn a file unwritable for anybody.

   Land the current permission with 0o7555 to remove the writable flag to
   anybody.
   Note: if the file was already writable for others or the group, we
   suppress this permission. *)
;;

val set_file_permissions : Unix.file_perm -> Path.file_name -> unit
(** [set_file_permissions perm fname] sets the file permissions of file
    [fname] to [perm]. *)
;;

(** {6 Primitives on channels cautiously dealing with [sdin] and [stdout].} *)

(** stdout and stdin are (conventionally) associated to the file name
   "-". We redefine open, close, and remove to implement this convention. *)

val remove : Path.file_name -> unit
(** Removing a file: we ensure the file is writable before removing it. *)
;;

val rename : Path.file_name -> Path.file_name -> unit
(** Rename a file into another one.
  Nothing happens when arguments are equal.
  Raise [Failure] when one (and only one) of the arguments is a standard
  channel. *)
;;

val open_out : Unix.file_perm -> Path.file_name -> out_channel
(** Open out for writing: we ensure the file is writable. *)
;;

val close_out : Path.file_name -> out_channel -> unit
(** Close an [out_channel] without closing [stdout]. *)
;;

val open_in : Path.file_name -> in_channel
(** Opening an [in_channel] without trying to reopen [stdin]. *)
;;

val close_in : Path.file_name -> in_channel -> unit
(** Close an [in_channel] without closing [stdin]. *)
;;

(** {6 Dealing with file name extensions.} *)

val chop_extension : Path.file_name -> string
(** Remove the extension of a file name if any. *)
;;

val add_extension : file_extension -> Path.file_name -> Path.file_name
(** Add the given extension to a file name if not a standard channel. *)
;;

val temporary_target :
  file_extension -> Path.file_name -> Path.file_name -> Path.file_name
(** [File.temporary_target ext src_name tgt_name] return a temporary file name
  suitable to run a file processor from [src_name] to [tgt_name], preserving
  [src_name] file integrity.
  The result can be [tgt_name], or even [src_name], provided that [src_name]
  file integrity is preserved; otherwise, it is a new [file_name] with extension
  [ext]. *)
;;

val remove_temporary_target : Path.file_name -> Path.file_name -> unit
(** Remove the temporary file name [temp_file_name] when it is not the
  [tgt_name] file. *)
;;
